package net.knarcraft.playerpayouts.config.payout;

import org.apache.commons.lang.NotImplementedException;
import org.jetbrains.annotations.NotNull;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

/**
 * A parser for the payout mode string
 */
public class PayoutActionParser {

    /**
     * Matches a payout component from the input string
     *
     * @param toMatch <p>The string to match</p>
     * @return <p>A payout component describing how to handle payouts</p>
     * @throws ParseException <p>If unable to parse the input string</p>
     */
    public static PayoutComponent matchPayoutComponent(@NotNull String toMatch) throws ParseException {
        List<String> parts = tokenize(toMatch);
        return parsePayoutComponent(parts, 0);
    }

    /**
     * Parses a payout component from the given input
     *
     * @param parts <p>The list of components to parse</p>
     * @param index <p>The index of the list to start at</p>
     * @return <p>The parsed payout component</p>
     * @throws ParseException <p>If unable to parse the input</p>
     */
    private static @NotNull PayoutComponent parsePayoutComponent(@NotNull List<String> parts,
                                                                 int index) throws ParseException {
        PayoutComponent payoutComponent = null;

        if (parts.size() <= index) {
            throw new ParseException("Found no payout mode information to parse", 0);
        }

        for (int i = index; i < parts.size(); i++) {
            /* If the payout component isn't null, we're in the next step of the for loop, and the entire value of 
               payoutComponent is the "left" component */
            if (payoutComponent == null) {
                String part = parts.get(i);
                PayoutTarget target = PayoutTarget.match(part);
                if (target == null) {
                    throw new ParseException("Unable to parse payout mode string component " + part + "!", i);
                }
                payoutComponent = target;
                if (parts.size() == 1) {
                    return payoutComponent;
                }
            }

            // Stop here, as the last component has already been parsed
            if (i + 1 >= parts.size()) {
                return payoutComponent;
            }

            // Parse the delimiter
            PayoutDelimiter delimiter = PayoutDelimiter.match(parts.get(++i).charAt(0));
            if (delimiter == null) {
                throw new ParseException("Unable to parse payout mode string delimiter " + parts.get(i) + "!", i);
            }

            if (delimiter == PayoutDelimiter.ADD) {
                PayoutTarget target2 = PayoutTarget.match(parts.get(++i));
                if (target2 == null) {
                    throw new ParseException("Unable to parse payout mode string component " + parts.get(i) + "!", i);
                }
                payoutComponent = new PayoutAction(payoutComponent, delimiter, target2);
                // The index has to be held back to prevent the parser from skipping the delimiter
                i--;
            } else if (delimiter == PayoutDelimiter.OR) {
                return new PayoutAction(payoutComponent, delimiter, parsePayoutComponent(parts, i + 1));
            } else {
                throw new NotImplementedException("That payout delimiter has not been implemented! Alert the developer!");
            }
        }

        return payoutComponent;
    }

    /**
     * Tokenizes a string
     *
     * @param input <p>A string.</p>
     * @return <p>A list of tokens.</p>
     */
    private static @NotNull List<String> tokenize(@NotNull String input) {
        List<String> tokens = new ArrayList<>();
        StringBuilder currentToken = new StringBuilder();
        for (int index = 0; index < input.length(); index++) {
            char character = input.charAt(index);
            if (PayoutDelimiter.match(character) != null) {
                // Tokenize delimiters, and tokenize everything read before the delimiter
                if (!currentToken.isEmpty()) {
                    tokens.add(currentToken.toString());
                }
                tokens.add(String.valueOf(character));
                currentToken = new StringBuilder();
            } else if (Character.isLetter(character)) {
                // Tokenize one letter
                tokenizeLetter(currentToken, character, input.length(), index, tokens);
            }

            // Unrecognized tokens are ignored, to allow for spaces
        }
        return tokens;
    }

    /**
     * Adds a letter to the token. Adds the current token to tokens if at the end of the input
     *
     * @param currentToken <p>The string builder containing the current token.</p>
     * @param character    <p>The character found in the input.</p>
     * @param inputLength  <p>The length of the given input</p>
     * @param index        <p>The index of the read character.</p>
     * @param tokens       <p>The list of processed tokens.</p>
     */
    private static void tokenizeLetter(@NotNull StringBuilder currentToken, char character, int inputLength,
                                       int index, @NotNull List<String> tokens) {
        currentToken.append(character);
        if (index == inputLength - 1) {
            tokens.add(currentToken.toString());
        }
    }

}
